--[[-------------------------------------------------------------------------
-- Simple Taunt Announce
--
-- Copyright 2011-2013 BeathsCurse (Saphod - Draenor EU)
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
-- http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
-------------------------------------------------------------------------]]--

local addonName, addon = ...
addon.version = GetAddOnMetadata(addonName, 'Version')

-- Expose our addon table through a global variable
_G[addonName] = addon

-- Configuration table with default values
-- Note: Valid values for own and other are: off, self, say, party, raid,
-- and instance (which work depends on status).
addon.cfg = {
	cfgversion = 1,
	isEnabled = true,
	-- When solo, how to announce your own taunts and those of others
	solo = { own = 'self', other = 'off' },
	-- When in an instance group
	instance = { own = 'self', other = 'self' },
	-- When in a party
	party = { own = 'self', other = 'self' },
	-- When in a raid
	raid = { own = 'say', other = 'self' },
}

-- Create frame for handling events
addon.frame = CreateFrame('Frame', addonName .. 'Frame')
addon.frame:Hide()

-- Local shorthands for accessing the configuration and frame
local cfg = addon.cfg
local frame = addon.frame

-- Function to enable addon
function addon:Enable()
	if not self.cfg.isEnabled then
		self.cfg.isEnabled = true
		self.frame:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end
end

-- Function to disable addon
function addon:Disable()
	if self.cfg.isEnabled then
		self.cfg.isEnabled = false
		self.frame:UnregisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end
end

-- Table of taunt aura IDs
-- Note: Death Grip, Provoke, and Leer of the Ox have different spell IDs
-- for the spell cast and the applied aura. The Shaman Earth Elemental AoE
-- taunt, Angered Earth (36213), does not apply an aura, so I left it out
-- to avoid immune spam on bosses.
local tauntAuras = {
	[355]=1,	-- Warrior, Taunt
	[2649]=1,	-- Hunter pet, Growl
	[6795]=1,	-- Druid, Growl
	[17735]=1,	-- Warlock Voidwalker, Suffering
	[20736]=1,	-- Hunter, Distracting Shot
	[49560]=1,	-- Death Knight, Death Grip (spell 49576)
	[56222]=1,	-- Death Knight, Dark Command
	[62124]=1,	-- Paladin, Hand of Reckoning
	[73684]=1,	-- Shaman, Unleash Earth
	[97827]=1,	-- Warlock, Provocation (Dark Apotheosis)
	[114198]=1,	-- Warrior, Mocking Banner
	[116189]=1,	-- Monk, Provoke (spell 115546)
	[118585]=1,	-- Monk, Leer of the Ox (glyph, spell 115543)
	[118635]=1,	-- Monk, AoE Provoke through Black Ox Statue (spell 115546)
	[123588]=1,	-- Hunter, Distracting Shot (glyphed, spell 20736)
}

-- Table of taunt spell IDs
-- Note: The Shaman Unleash Earth taunt is applied with Unleash Elements, so
-- we cannot check against the spell without triggering on other types.
local tauntSpells = {
	[355]=1,	-- Warrior, Taunt
	[2649]=1,	-- Hunter pet, Growl
	[6795]=1,	-- Druid, Growl
	[17735]=1,	-- Warlock Voidwalker, Suffering
	[20736]=1,	-- Hunter, Distracting Shot
	[49576]=1,	-- Death Knight, Death Grip (aura 49560)
	[56222]=1,	-- Death Knight, Dark Command
	[62124]=1,	-- Paladin, Hand of Reckoning
	[97827]=1,	-- Warlock, Provocation (Dark Apotheosis)
	[114198]=1,	-- Warrior, Mocking Banner
	[115543]=1,	-- Monk, Leer of the Ox (glyph, aura 118585)
	[115546]=1,	-- Monk, Provoke (auras 116189 and 118635)
}

-- Variables used to prevent flooding on AoE taunts
local lastTime, lastSpellID
local mockingBanners = {}

-- Local names for globals used in CLEU handler
local GetSpellLink = GetSpellLink
local GetZonePVPInfo = GetZonePVPInfo
local IsInGroup = IsInGroup
local IsInRaid = IsInRaid
local UnitClass = UnitClass
local UnitCreatureFamily = UnitCreatureFamily
local UnitInBattleground = UnitInBattleground
local UnitName = UnitName

-- Update configuration table
-- Note: Assumes all keys are simple and values are tables or simple.
-- Will overwrite a non-table with a table, but not the other way around.
local function UpdateCfg(dst, src)
	for k, v in pairs(src) do
		if type(v) == 'table' then
			if type(dst[k]) == 'table' then
				-- Update a table with a table
				UpdateCfg(dst[k], v)
			else
				-- Overwrite a value with a table
				dst[k] = UpdateCfg({}, v)
			end
		elseif type(dst[k]) ~= 'table' then
			-- Update a value with a value
			dst[k] = v
		end
	end
	return dst
end

-- Split a string into whitespace delimited parts and return them as a list
-- Note: Simple and slow, but we only use it to parse the slash commands.
local function SplitString(s)
	local t = {}
	-- Loop over matches of non-whitespace sequences in s, adding them to t
	for w in s:gmatch('%S+') do
		t[#t + 1] = w
	end
	return unpack(t)
end

-- Get name of player
-- Note: We cannot read the name into a local variable on load since
-- UnitName may return UNKNOWNOBJECT until the player has fully loaded.
-- We could potentially use the GUID instead.
local GetPlayerName
do
	local playerName

	function GetPlayerName()
		if not playerName then
			local a = UnitName('player')
			if a and a ~= UNKNOWNOBJECT then
				playerName = a
			end
		end
		return playerName
	end
end

-- Get spell link for a spellID
local GetSpellLinkCached
do
	local spellLinkCache = {}

	function GetSpellLinkCached(spellID)
		local a = spellLinkCache[spellID]
		if not a then
			a = GetSpellLink(spellID)
			spellLinkCache[spellID] = a
		end
		return a
	end
end

-- Event dispatcher
frame:SetScript('OnEvent', function(self, event, ...)
	local a = self[event]
	if a then
		a(self, ...)
	end
end)

-- Perform actions once all saved variables are loaded
function frame:ADDON_LOADED(name)
	if name ~= addonName then return end

	-- We do not have to be called for any further addons
	self:UnregisterEvent('ADDON_LOADED')

	-- Update current configuration table with saved
	if not STACFG then STACFG = {} end
	UpdateCfg(cfg, STACFG)

	-- Register combat log event handling if announcing is on
	if cfg.isEnabled then
		self:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
	end
end

-- Perform actions as player is about to enter world
function frame:PLAYER_LOGIN()
	-- Add slash command
	SLASH_STA1 = '/sta'
	SlashCmdList['STA'] = function(msg)
		self:OnSlashCmd(SplitString((msg or ''):lower()))
	end
end

-- Perform actions as player is about to leave world
function frame:PLAYER_LOGOUT()
	-- Update saved configuration table with current
	UpdateCfg(STACFG, cfg)
end

-- The actual combat log event handler
-- Note: Naming the args was faster than select, and Lua adjusts the number
-- of arguments automatically.
function frame:COMBAT_LOG_EVENT_UNFILTERED(timeStamp, subEvent, _, _, sourceName, sourceFlags, _, _, destName, _, _, spellID, _, _, missType)
	local failString

--[[
	-- For debugging purposes
	if subEvent == 'SPELL_CAST_SUCCESS' then
		print((sourceName or '?') .. ' -> ' .. (destName or '?') .. ' spellcast ' .. GetSpellLinkCached(spellID) .. ' with ID ' .. spellID .. ' flags ' .. bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_MASK))
	elseif subEvent == 'SPELL_AURA_APPLIED' then
		print((sourceName or '?') .. ' -> ' .. (destName or '?') ..  ' aura ' .. GetSpellLinkCached(spellID) .. ' with ID ' .. spellID .. ' flags ' .. bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_MASK))
	end
--]]

	-- Check if event was a taunt aura applied or taunt spell missed
	-- Note: We could use SPELL_CAST_SUCCESS instead, but it has a few issues,
	-- like the Shaman taunt being cast through Unleash Elements. Also,
	-- casting an AoE taunt does not mean it actually taunts anything.
	-- Checking for the aura will be slightly slower though, because there
	-- are more of those events.
	if subEvent == 'SPELL_AURA_APPLIED' then
		-- If applied aura was not a taunt, return
		if not tauntAuras[spellID] then return end
	elseif subEvent == 'SPELL_MISSED' then
		-- If missed spell was not a taunt, return
		if not tauntSpells[spellID] then return end
		-- Set failString to a suitable indication the taunt failed
		failString = ' failed (' .. (missType or '?'):lower() .. ')'
	else
		-- Neither an aura applied nor a spell missed, so return
		return
	end

	-- If player is in a battleground or an arena, return
	if UnitInBattleground('player') or GetZonePVPInfo() == 'arena' then return end

	-- Check if time and ID was same as last
	-- Note: This is to prevent flooding announcements on AoE taunts.
	if timeStamp == lastTime and spellID == lastSpellID then return end

	-- Update last time and ID
	lastTime, lastSpellID = timeStamp, spellID

	-- Special handling of Mocking Banner
	-- Note: Since mocking banner potentially taunts every second for 30
	-- seconds, we limit the announcing to the first taunt for each player.
	if spellID == 114198 then
		-- Get time of last Mocking Banner taunt announced for this source
		local lastMB = mockingBanners[sourceName]
		local currentMB = time()

		-- If we already announced a Mocking Banner taunt for this source
		-- within the last 40 seconds, return
		-- Note: While Mocking Banner only lasts 30 seconds, the cooldown
		-- is 3 minutes, so we check for a liberal 40 second window.
		if lastMB and currentMB <= lastMB + 40 then
			return
		end

		-- Add, or update, the last Mocking Banner taunt for this source
		mockingBanners[sourceName] = currentMB
	end

	-- Figure out grouping status
	-- Note: Passing LE_PARTY_CATEGORY_INSTANCE to IsInGroup() and IsInRaid()
	-- appears to check if we are in a group assembled through the instance
	-- finder (LFG/LFR/BG/etc.), where the chat channel is now INSTANCE_CHAT.
	-- Passing LE_PARTY_CATEGORY_HOME checks if we are in a normal group. If
	-- you join the instance finder as a group, you will be in both the
	-- original group/raid and the instance group. Passing nothing checks for
	-- either.
	local inInstance = IsInGroup(LE_PARTY_CATEGORY_INSTANCE)
	local inGroup, inRaid = IsInGroup(), IsInRaid()

	local config

	-- Select config based on grouping status
	if inInstance then
		config = cfg.instance
	elseif inRaid then
		config = cfg.raid
	elseif inGroup then
		config = cfg.party
	else
		config = cfg.solo
	end

	-- Check if source was the player, player's pet, or other
	if sourceName == GetPlayerName() then
		-- If own is off for this config, return
		if config.own == 'off' then return end

		-- Announce taunt
		self:AnnounceTaunt(config.own, string.format('Taunting %s with %s%s', destName or '?', GetSpellLinkCached(spellID), failString or ''))
	elseif bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_MINE) > 0 then
		-- If own is off for this config, return
		if config.own == 'off' then return end

		-- Announce taunt
		self:AnnounceTaunt(config.own, string.format('My %s taunting %s with %s%s', sourceName or '?', destName or '?', GetSpellLinkCached(spellID), failString or ''))
	else
		-- If other is off for this config, return
		if config.other == 'off' then return end

		-- Announce if source was a player/pet in our party/raid
		if (inRaid and bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_RAID) > 0) or (inGroup and bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_PARTY) > 0) then
			local name = sourceName or '?'

			-- Apply class color to name if announcing to self and source was a player
			if config.other == 'self' and bit.band(sourceFlags, COMBATLOG_OBJECT_TYPE_PLAYER) > 0 then
				local _, classFileName = UnitClass(name)
				name = '|c' .. RAID_CLASS_COLORS[classFileName or 'PRIEST'].colorStr .. name .. '|r'
			end

			self:AnnounceTaunt(config.other, string.format('%s taunting %s with %s%s', name, destName or '?', GetSpellLinkCached(spellID), failString or ''))
		end
	end
end

-- Register the events we handle
frame:RegisterEvent('ADDON_LOADED')
frame:RegisterEvent('PLAYER_LOGIN')
frame:RegisterEvent('PLAYER_LOGOUT')

-- Print with STA prefix
function frame:Print(...)
	return print('|cFFF0F050STA|r:', ...)
end

-- Formatted print with STA prefix
function frame:Printf(s, ...)
	return print('|cFFF0F050STA|r:', string.format(s, ...))
end

-- Announce message to self or channel based on mode
-- Note: mode has to be self, say, party, raid, or instance
function frame:AnnounceTaunt(mode, msg)
	if mode == 'self' then
		self:Print(msg)
	else
		-- If we are announcing to instance, change mode to INSTANCE_CHAT.
		if (mode == 'instance') then mode = 'INSTANCE_CHAT' end

		-- Here mode is say, party, raid, or INSTANCE_CHAT
		SendChatMessage(msg, mode:upper())
	end
end

-- Print a line with own and other modes in config
function frame:PrintMode(mode, config)
	self:Printf('|cFF20E020%s|r modes: own -> |cFF20E020%s|r, other -> |cFF20E020%s|r', mode, config.own, config.other)
end

-- Update config with the modes from own and other
function frame:ParseCmdModes(cmd, config, own, other)
	-- Update own mode
	if own then
		if own == 'off' then
			config.own = 'off'
		elseif own == 'self' then
			config.own = 'self'
		elseif own == 'say' then
			config.own = 'say'
		elseif own == 'instance' and cmd == 'instance' then
			config.own = 'instance'
		elseif own == 'party' and (cmd == 'party' or cmd == 'raid') then
			config.own = 'party'
		elseif own == 'raid' and cmd == 'raid' then
			config.own = 'raid'
		else
			self:Print('invalid mode for |cFF20E020' .. cmd .. '|r own:', own)
			return
		end
	end

	-- Update other mode
	if other then
		if other == 'off' then
			config.other = 'off'
		elseif other == 'self' then
			config.other = 'self'
		elseif other == 'say' then
			config.other = 'say'
		elseif other == 'instance' and cmd == 'instance' then
			config.other = 'instance'
		elseif other == 'party' and (cmd == 'party' or cmd == 'raid') then
			config.other = 'party'
		elseif other == 'raid' and cmd == 'raid' then
			config.other = 'raid'
		else
			self:Print('invalid mode for |cFF20E020' .. cmd .. '|r other:', other)
			return
		end
	end
end

-- Slash command handler
function frame:OnSlashCmd(cmd, ...)
	if cmd == 'on' then
		-- Register combat log event handling to start announcing
		self:Print('Announce turned |cFF20E020on|r')
		addon:Enable()
	elseif cmd == 'off' then
		-- Unregister combat log event handling to stop announcing
		self:Print('Announce turned |cFFE02020off|r')
		addon:Disable()
	elseif cmd == 'solo' then
		self:ParseCmdModes(cmd, cfg.solo, ...)
		self:PrintMode(cmd, cfg.solo)
	elseif cmd == 'instance' then
		self:ParseCmdModes(cmd, cfg.instance, ...)
		self:PrintMode(cmd, cfg.instance)
	elseif cmd == 'party' then
		self:ParseCmdModes(cmd, cfg.party, ...)
		self:PrintMode(cmd, cfg.party)
	elseif cmd == 'raid' then
		self:ParseCmdModes(cmd, cfg.raid, ...)
		self:PrintMode(cmd, cfg.raid)
	elseif cmd == 'info' then
		-- Show a little profiling info
		UpdateAddOnMemoryUsage()
		local usedKB = GetAddOnMemoryUsage(addonName)
		self:Printf('Mem usage %.2f KB', usedKB)
		if GetCVar('scriptProfile') == '1' then
			UpdateAddOnCPUUsage()
			local addOnTime = GetAddOnCPUUsage(addonName)
			self:Printf('AddOn CPU usage %.2f (%.2f%%)', addOnTime, (100 * addOnTime) / GetScriptCPUUsage())
			local frameTime, frameCount = GetFrameCPUUsage(frame)
			self:Printf('Frame CPU usage %.2f in %d calls (%.4f per call)', frameTime, frameCount, frameTime / frameCount)
		end
	elseif cmd == 'help' then
		-- Show an explanation of status and modes
		self:Print('For each status (|cFF20E020solo instance party|r and |cFF20E020raid|r), own and')
		self:Print('other mode can be set to |cFF20E020off self say instance party|r or |cFF20E020raid|r.')
		self:Print('Own covers taunts made by you or your pet, other covers')
		self:Print('taunts made by any player or pet in your party/raid.')
		self:Print('Examples:')
		self:Print('  |cFF20E020/sta party|r - print current party modes')
		self:Print('  |cFF20E020/sta solo self|r - when solo, own taunts to yourself')
		self:Print('  |cFF20E020/sta raid say self|r - when raid, own to say, others to self')
	else
		-- By default, show status and options
		self:Print(addonName, addon.version)
		self:Print('Syntax: /sta <command> [options]')
		self:Print('  |cFF20E020on/off|r - turn announcing on/off')
		self:Print('  |cFF20E020solo/instance/party/raid [own [other]]|r - set mode')
		self:Print('  |cFF20E020help|r - show more help')
		if cfg.isEnabled then
			self:Print('Announcing is |cFF20E020on|r')
		else
			self:Print('Announcing is |cFFE02020off|r')
		end
	end
end
